#ifndef PBP_CUSTOM_DENOISE_POTENTIAL_HPP
#define PBP_CUSTOM_DENOISE_POTENTIAL_HPP

#include <stdexcept>
#include "math.h"
#include <Interval.hpp>
#include <GMPotential.hpp>

#define sqr(a) ((a)*(a))

class DenoisePotential : public pbp::GMPotential {
private:
  const pbp::Array<double> img;
  const pbp::Array<double> alpha;
public:
  DenoisePotential(const pbp::Array<double> img, const pbp::Array<double> alpha) : img(img), alpha(alpha) {};
  virtual pbp::Energy nodePotential(pbp::Label x, pbp::Index n) const {
    return alpha[0]*sqr(x[0]-img[n]);
  };
  virtual pbp::Interval nodeInterval(pbp::Label x, pbp::Index n, pbp::Index dimension, pbp::Energy U) const {
    double left, right;
    pbp::Interval bounds;
    switch (dimension) {
      case 0: 
        if ((U < 0.0)) {
          return pbp::Interval();
        } else
        if ((0.0 <= U)) {
          left = img[n]-sqrt(U/alpha[0]);
          right = img[n]+sqrt(U/alpha[0]);
          bounds.set_union(left, right);
        } else
          return pbp::Interval();
        break;
      default: throw std::runtime_error("GMPotential: currDim > 0.");
    }
    return bounds;
  };
  virtual pbp::Energy edgePotential(pbp::Label x, pbp::Label y, pbp::Index e) const {
    return ((alpha[2] < sqr(x[0]-y[0])) ? (alpha[1]*alpha[2]) : (alpha[1]*sqr(x[0]-y[0])));
  };
  virtual pbp::Interval edgeXInterval(pbp::Label x, pbp::Label y, pbp::Index e, pbp::Index dimension, pbp::Energy U) const {
    double left, right;
    pbp::Interval bounds;
    switch (dimension) {
      case 0: 
        if ((U < 0.0)) {
          return pbp::Interval();
        } else
        if ((alpha[1]*alpha[2] <= U)) {
          left = y[0]+sqrt(alpha[2]);
          right = INFINITY;
          bounds.set_union(left, right);
          left = -INFINITY;
          right = y[0]-sqrt(alpha[2]);
          bounds.set_union(left, right);
          left = y[0]-sqrt(U/alpha[1]);
          right = y[0]+sqrt(U/alpha[1]);
          bounds.set_union(left, right);
        } else
        if (((0.0 <= U) && (U < alpha[1]*alpha[2]))) {
          left = y[0]-sqrt(alpha[2]);
          right = y[0]+sqrt(alpha[2]);
          bounds.set_union(left, right);
          left = y[0]-sqrt(U/alpha[1]);
          right = y[0]+sqrt(U/alpha[1]);
          bounds.set_union(left, right);
        } else
          return pbp::Interval();
        break;
      default: throw std::runtime_error("GMPotential: currDim > 0.");
    }
    return bounds;
  };
  virtual pbp::Interval edgeYInterval(pbp::Label x, pbp::Label y, pbp::Index e, pbp::Index dimension, pbp::Energy U) const {
    double left, right;
    pbp::Interval bounds;
    switch (dimension) {
      case 0: 
        if ((U < 0.0)) {
          return pbp::Interval();
        } else
        if ((alpha[1]*alpha[2] <= U)) {
          left = x[0]+sqrt(alpha[2]);
          right = INFINITY;
          bounds.set_union(left, right);
          left = -INFINITY;
          right = x[0]-sqrt(alpha[2]);
          bounds.set_union(left, right);
          left = x[0]-sqrt(U/alpha[1]);
          right = x[0]+sqrt(U/alpha[1]);
          bounds.set_union(left, right);
        } else
        if (((0.0 <= U) && (U < alpha[1]*alpha[2]))) {
          left = x[0]-sqrt(alpha[2]);
          right = x[0]+sqrt(alpha[2]);
          bounds.set_union(left, right);
          left = x[0]-sqrt(U/alpha[1]);
          right = x[0]+sqrt(U/alpha[1]);
          bounds.set_union(left, right);
        } else
          return pbp::Interval();
        break;
      default: throw std::runtime_error("GMPotential: currDim > 0.");
    }
    return bounds;
  };
};

#endif
