/* pbpTypes.hpp
 * 
 * Copyright: Oliver Müller (TNT, Leibniz Universität Hannover), 2013
 * E-Mail: omueller@tnt.uni-hannover.de
 */
#ifndef PBP_PBP_TYPES_HPP
#define PBP_PBP_TYPES_HPP

#include <stdexcept>
#include <iostream>
#include <vector>
#include <string>
#include <boost/shared_ptr.hpp>

namespace pbp {

  template<class T> class Array; // Forward declaration.

  typedef unsigned int Index;
  typedef double Energy;
  typedef Array<double> Label;

  template<class T>
  class Array {
  public:
    typedef std::size_t size_type;
    typedef T* iterator;
    typedef const T* const_iterator;

    Array(const T *firstPtr, const T *endPtr) : firstPtr(const_cast<T*>(firstPtr)), endPtr(const_cast<T*>(endPtr)) {};
    
    Array(const T *firstPtr, const T *endPtr, const std::string &name) : firstPtr(const_cast<T*>(firstPtr)), endPtr(const_cast<T*>(endPtr)), name(name) {};

    Array(const Array &array) : firstPtr(array.firstPtr), endPtr(array.endPtr), vector(array.vector), name(array.name) {};

    Array(const std::vector<T> &v) : vector(new std::vector<T>(v)) {
      assign(&(*vector->begin()), &(*vector->end()));
    };

    Array() : firstPtr(NULL), endPtr(NULL) {};

    Array(const std::string &name) : firstPtr(NULL), endPtr(NULL), name(name) {};

    operator T*() {
      return firstPtr;
    };

    operator const T*() const {
      return firstPtr;
    };

    T* begin() { return firstPtr; };
    const T* begin() const { return firstPtr; };

    T* end() { return endPtr; };
    const T* end() const { return endPtr; };

    bool empty() const {
      return firstPtr == endPtr;
    };

    size_type size() const { return endPtr - firstPtr; };

    inline T& operator[](const unsigned int &index) {
      if (firstPtr + index >= endPtr)
      {
        #pragma omp critical (Array)
          std::cout << "Index out of range in Array " << name << "(" << size() << ")[" << index << "]" << std::endl;
        throw std::out_of_range("Index out of range.");
      }
      return *(firstPtr + index);
    }
    inline const T& operator[](const unsigned int &index) const {
      if (firstPtr + index >= endPtr)
      {
        #pragma omp critical (Array)
          std::cout << "Index out of range in Array " << name << "(" << size() << ")[" << index << "]" << std::endl;
        throw std::out_of_range("Index out of range.");
      }
      return *(firstPtr + index);
    }
  protected:
    void assign(T *firstRef, T *endRef) {
      firstPtr = firstRef;
      endPtr = endRef;
    }
  private:
    boost::shared_ptr<std::vector<T> > vector;
    T *firstPtr;
    T *endPtr;
    std::string name;
  };

  /* makeView methods: */

  template <typename T>
  Array<T> makeView(Array<T> &v) {
    return Array<T>(v);
  };

  template <typename T>
  Array<T> makeView(std::vector<T> &v) {
    return Array<T>(&(*v.begin()), &(*v.end()));
  };

  template <typename T>
  Array<Array<T> > makeView(std::vector<std::vector<T> > &v) {
    Array<Array<T> > vv( std::vector<Array<T> >(v.size()) );
    typename std::vector<std::vector<T> >::iterator vit = v.begin();
    for (typename Array<Array<T> >::iterator it = vv.begin(); it != vv.end(); it++, vit++) {
      *it = makeView(*vit);
    }
    return vv;
  };

  /* makeCopy methods: */

  template <typename T>
  Array<T> makeCopy(const Array<T> &v) {
    return Array<T>(std::vector<T>(v.begin(), v.end()));
  };

  template <typename T>
  Array<Array<T> > makeCopy(const Array<Array<T> > &v) {
    Array<Array<T> > vv( std::vector<Array<T> >(v.size()) );
    typename Array<Array<T> >::const_iterator vit = v.begin();
    for (typename Array<Array<T> >::iterator it = vv.begin(); it != vv.end(); it++, vit++) {
      *it = makeCopy(*vit);
    }
    return vv;
  };

  template <typename T>
  Array<T> makeCopy(const std::vector<T> &v) {
    return Array<T>(v);
  };

  template <typename T>
  Array<Array<T> > makeCopy(const std::vector<std::vector<T> > &v) {
    Array<Array<T> > vv( std::vector<Array<T> >(v.size()) );
    typename std::vector<std::vector<T> >::const_iterator vit = v.begin();
    for (typename Array<Array<T> >::iterator it = vv.begin(); it != vv.end(); it++, vit++) {
      *it = makeCopy(*vit);
    }
    return vv;
  };

  /* Stream output methods: */

  template<class T>
  std::ostream& writeVector(std::ostream& os, const T& vector) {
    if (vector.empty())
    {
      os << "[]";
    } else
    {
      if (vector.size() >= 4)
        os << "(" << vector.size() << ")[";
      else
        os << "[";
      unsigned int k = 0;
      for (typename T::const_iterator it = vector.begin(); it != vector.end(); ) {
        if (k >= 10) {
          os << ", ...]";
          break;
        }
        os << *it;
        if (++it != vector.end()) {
          os << ", ";
        } else
        {
          os << "]";
        }
        k++;
      }
    }
    return os;
  }

  template<class T> 
  std::ostream& operator<<(std::ostream& os, const pbp::Array<T>& vector) {
    return writeVector(os, vector);
  }

  template<class T> 
  std::ostream& operator<<(std::ostream& os, const std::vector<T>& vector) {
    return writeVector(os, vector);
  }

}
#endif
