% function generateGMmex(mexName, potentialHeaderFileName, className, parameters)
% 
% mexName: Name of generated MEX file (without file extension)
% potentialHeaderFileName: Filename of header containing GMpotential class
%                          implementation (e.g. generated using
%                          generateGMPotentialHeader)
% className: Class name of GMpotential implementation
% parameters: Cell array of input variables used by GMpotential
%             implementation
% 
% Usage example:
% >> generateGMPotentialHeader('foo.hpp', 'fooGMpot', sym('x^2'), sym('(x-y)^2'));
% >> generateGMmex('fooBP', 'foo.hpp', 'fooGMpot', []);
% >> compileGMmex('fooBP', '/usr/lib/pbp');
% 
% Copyright: Oliver Müller (TNT, Leibniz Universität Hannover), 2013
function generateGMmex(mexName, potentialHeaderFileName, className, parameters)

  GMPotenialHeaderInc = potentialHeaderFileName;
  numRequiredOptArgs = numel(parameters);

  if numel(parameters) <= 0
    potParamsStr = '';
  else
    potParamsStr = parameterToStr(parameters{1}, 1);
    for p=2:numel(parameters)
      potParamsStr = [potParamsStr, ', ', parameterToStr(parameters{p}, p)]; %#ok<AGROW>
    end
    potParamsStr = ['(', potParamsStr, ')'];
  end

  mexTemplate = [...
    '#include "mex.h"\n', ...
    '#include <mexPBP.hpp>\n', ...
    '#include <%s>\n', ...
    '\n', ...
    '#define NUM_REQUIRED_OPT_ARGS %d\n', ...
    '\n', ...
    'void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {\n', ...
    '  mexPBPvalidateParams(nlhs, plhs, nrhs, prhs, NUM_REQUIRED_OPT_ARGS);\n', ...
    '  %s pot%s;\n', ...
    '  mexPBPrun(plhs, prhs, pot);\n', ...
    '}\n', ...
    ];

  % Write everything to file:
  sourceFile = [mexName, '.cpp'];

  text = sprintf(mexTemplate, GMPotenialHeaderInc, numRequiredOptArgs, ...
    className, potParamsStr);
  
  writeToFileWithBackup(sourceFile, text);

end

function writeToFileWithBackup(FileName, text)

  if exist(FileName, 'file')
    fileInfo = dir(FileName);
    if ~isempty(fileInfo)
      fileSize = fileInfo.bytes;
      if fileSize ~= numel(text)
        movefile(FileName, getBackupFileName(FileName));
      else
        oldText = fileread(FileName);
        if any(oldText(:) ~= text(:))
          movefile(FileName, getBackupFileName(FileName));
        end
      end
    end
  end

  fID = fopen(FileName, 'w');
  fprintf(fID, text);
  fclose(fID);

end

function backupFileName = getBackupFileName(FileName)
  backupFileName = sprintf('%s.backup', FileName);
  nBackup = 1;
  while exist(backupFileName, 'file')
    backupFileName = sprintf('%s.backup%d', FileName, nBackup);
    nBackup = nBackup + 1;
  end
end

function str = parameterToStr(parameter, id)
  str = sprintf('prhs[NUM_REQUIRED_ARGS+%d]', id-1);
  if ~isParameterMxArray(parameter)
    str = sprintf('mexPBPtoVector(%s, "%s")', str, getParameterName(parameter));
  end
end

function bool = isParameterMxArray(param)
  if (numel(param) > 0) && (param(end) == '*')
    bool = true;
  else
    bool = false;
  end
end

function str = getParameterName(param)
  id = strfind(param, '*');
  if numel(id) >= 1
    str = param(1:min(id)-1);
  else
    str = param;
  end
end
